/*
 * Copyright 2022 KylinSoft Co., Ltd.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "multitaskviewmanager.h"

#include <KF5/KGlobalAccel/KGlobalAccel>
#include <KF5/KWindowSystem/KWindowSystem>

#include <libkysysinfo.h>

#include <QAction>
#include <QApplication>
#include <QQmlContext>
#include <QtDBus>
#include <QTranslator>
#include <QLocale>

#include "desktopbackground.h"
#include "icontheme.h"

namespace MultitaskView {

bool MultitaskViewManager::useOpenGL_ = true;

MultitaskViewManager::MultitaskViewManager()
{
    setProjectType();

    registerQmlTypes();

    registerShortcut();

    initDBusService();

    installTranslator();
}

MultitaskViewManager::~MultitaskViewManager()
{
    dbusService_.reset();

    close();

    if (translator_)
    {
        translator_->deleteLater();
        translator_ = nullptr;
    }
}

void MultitaskViewManager::grabbedKeyboardEvent(QKeyEvent *e)
{
    if (view_.get()) {
        view_->contentItem()->setFocus(true);
        QCoreApplication::sendEvent(view_.get(), e);
    }
}

void MultitaskViewManager::open()
{
    qDebug() << "开启多任务视图" << __FUNCTION__;
    if (isOpen)
    {
        return;
    }

    //可旋转的机器，多任务视图打开的时候，禁用旋转
    if(projectType_ == MultitaskViewManager::TABLET && model_->isTabletMode())
    {
        if(getRotationModeStatus())
        {
            setRotationModeStatus(!isOpenRotationMode);
        }
    }

    initView();

    if (!keyboardGrab) {
        keyboardGrab = KWin::effects->grabKeyboard(this);
        KWin::effects->setActiveFullScreenEffect(this);
    }

    isOpen = true;
}

bool MultitaskViewManager::getRotationModeStatus()
{
    QDBusMessage message = QDBusMessage::createMethodCall("com.kylin.statusmanager.interface",
                                                          "/",
                                                          "com.kylin.statusmanager.interface",
                                                          "get_auto_rotation");

    QDBusMessage response = QDBusConnection::sessionBus().call(message);

    if (response.type() == QDBusMessage::ReplyMessage)
    {
        isOpenRotationMode = response.arguments().takeFirst().toBool();
    }
    else
    {
        // invalid response, keep default.
    }
    return isOpenRotationMode;
}

void MultitaskViewManager::setRotationModeStatus(bool rotationModeStatus)
{
    QDBusInterface iface("com.kylin.statusmanager.interface",
                         "/",
                         "com.kylin.statusmanager.interface");

    if (!iface.isValid()) {
        qDebug() << QDBusConnection::sessionBus().lastError().message();
    }

    QDBusReply<void> reply = iface.call("set_auto_rotation",rotationModeStatus,"","");

    if (reply.isValid()) {
        qDebug() << "set Rotation mode success";
        return;
    }

    qDebug() << "set Rotation mode failed" << reply.error();
}

void MultitaskViewManager::close()
{
    qDebug() << "关闭多任务视图";
    if (!isOpen)
    {
        return;
    }

    //可旋转的机器，多任务视图打开前 是打开了旋转的 那么关闭多任务视图的同时  打开旋转
    if(projectType_ == MultitaskViewManager::TABLET && model_->isTabletMode())
    {
        if(isOpenRotationMode)
        {
            setRotationModeStatus(isOpenRotationMode);
        }
    }

    if (keyboardGrab) {
        KWin::effects->ungrabKeyboard();
    }
    keyboardGrab = false;
    KWin::effects->setActiveFullScreenEffect(nullptr);

    destroyView();

    isOpen = false;
}

void MultitaskViewManager::showDesktop(bool show)
{
    KWin::effects->setShowingDesktop(show);
}

void MultitaskViewManager::initModel()
{
    qDebug() << "创建model数据" << __FUNCTION__;
    model_.reset(new MultitaskViewModel());
}

void MultitaskViewManager::setProjectType()
{
    if ( "V10SP1" == getProjectCodeName() && '\u0000' == getProductFeatures() )
    {
        projectType_ = MultitaskViewManager::MASTER;
    }
    if ( 3 == getProductFeatures() )
    {
        projectType_ = MultitaskViewManager::XC;
    }

}

void MultitaskViewManager::setQmlUrl()
{
    if (projectType_ == MultitaskViewManager::TABLET || (projectType_ == MultitaskViewManager::XC && model_->isTabletMode()))
    {
        qmlUrl_ = QUrl("qrc:/qml/TabletViewMain.qml");
        isQmlTabletMode = true;
    }
    else if (projectType_ == MultitaskViewManager::PC || (projectType_ == MultitaskViewManager::XC && !model_->isTabletMode()))
    {
        qmlUrl_ = QUrl("qrc:/qml/multitaskview.qml");
        isQmlTabletMode = false;
    }
    else if (projectType_ == MultitaskViewManager::MASTER)
    {
        qmlUrl_ = QUrl("qrc:/qml/MasterMultitaskView.qml");
        isQmlTabletMode = false;
    }

}

void MultitaskViewManager::registerQmlTypes()
{
    qmlRegisterType<DesktopBackground>("org.kde.kwin", 3, 0, "DesktopBackground");
    qmlRegisterType<IconTheme>("IconTheme", 1, 0, "IconTheme");
}

void MultitaskViewManager::registerShortcut()
{
    auto* action = new QAction(this);
    action->setObjectName("ShowMultitaskView");
    action->setText("Show Multitask View");

    QKeySequence keySeq(Qt::META + Qt::Key_Tab);
    KGlobalAccel::self()->setDefaultShortcut(action, QList<QKeySequence>() << keySeq);
    KGlobalAccel::self()->setShortcut(action, QList<QKeySequence>() << keySeq);

    QObject::connect(action, &QAction::triggered, this, &MultitaskViewManager::open);
}

void MultitaskViewManager::initDBusService()
{
    dbusService_.reset(new DBusService(this));
}

void MultitaskViewManager::installTranslator()
{
    translator_ = new QTranslator(qApp);

    QLocale locale = QLocale::system();

    if (locale.language() == QLocale::Chinese)
    {
        translator_->load(":/ukui-window-switch_zh_CN.qm");
    }
    else if (locale.language() == QLocale::Tibetan)
    {
        translator_->load(":/ukui-window-switch_bo_CN.qm");
    }

    QApplication::installTranslator(translator_);
}

void MultitaskViewManager::initView()
{
    qDebug() << " 初始化多任务视图 " << __FUNCTION__;
    view_.reset(new QQuickView());

    view_->setFlag(Qt::WindowType::BypassWindowManagerHint);

    KWindowSystem::setType(view_->winId(), NET::Dock);

    qDebug() << "[DEBUG]: multitask view uses " << view_->sceneGraphBackend();
    if (view_->rendererInterface()->graphicsApi() < 2 || !KWin::effects->isOpenGLCompositing())
    {
        useOpenGL_ = false;
    }

    initModel();

    view_->rootContext()->setContextProperty("multitaskViewManager", this);
    view_->rootContext()->setContextProperty("multitaskViewModel", model_.get());

    setQmlUrl();

    qDebug() << "加载的qml为 " << qmlUrl_;
    view_->setSource(qmlUrl_);

    // 必须在加载了QML之后才能设置view_对象的大小
    qDebug() << " 设置的多任务视图的位置在" << KWin::effects->virtualScreenGeometry();
    view_->setGeometry(KWin::effects->virtualScreenGeometry());

    connectSignals();

    qDebug() << "开始显示多任务视图";
    view_->show();

    view_->requestActivate();
}

void MultitaskViewManager::destroyView()
{
    if (view_ == nullptr)
    {
        return;
    }

    if (view_->isVisible())
    {
        view_->hide();
    }

    // 1. 由于可以在QML中通过JS调用到close函数，
    // 因此，不能直接销毁view_对象；必须延迟
    // 销毁该对象。
    // 2. model_对象可以立刻销毁。但是，为了对称
    // 和简化逻辑，需要保证view_对象和model_对象的
    // 互相依赖的关系。因此，同时延迟销毁model_对象。
    // 3. MultitaskViewManager的析构函数需要确保
    // view_对象和model_对象完成销毁。为了简化逻辑，
    // MultitaskViewManager的析构函数选择直接调用
    // destroyView函数，即采用了延迟销毁的做法。
    view_.release()->deleteLater();

    model_.release()->deleteLater();
}

void MultitaskViewManager::connectSignals()
{
    qDebug() << "创建信号" << __FUNCTION__;
    const auto* model = model_.get();
    const auto* rootObject = view_->rootObject();
    connect(model, SIGNAL(appWindowCreated(int)),
            rootObject, SIGNAL(qmlAppWindowCreated(int)));
    connect(model, SIGNAL(appWindowActivated()),
            rootObject, SIGNAL(qmlAppWindowActivated()));
    connect(model, SIGNAL(appWindowRemoved(int)),
            rootObject, SIGNAL(qmlAppWindowRemoved(int)));
    connect(model, SIGNAL(screenCountChanged(int, int)),
            rootObject, SIGNAL(qmlScreenCountChanged(int, int)));
    connect(model, SIGNAL(screenSizeChanged()),
            rootObject, SIGNAL(qmlScreenSizeChanged()));
    connect(model, SIGNAL(tabletModeChanged(bool)),
            rootObject, SIGNAL(qmlDeviceModeChanged(bool)));
    if (!isQmlTabletMode)
    {
        connect(model, SIGNAL(appWindowDesktopChanged(int, int, int)),
                rootObject, SIGNAL(qmlAppWindowDesktopChanged(int, int, int)));
        connect(model, SIGNAL(desktopRemoved(int)),
                rootObject, SIGNAL(qmlDesktopRemoved(int)));
        connect(model, SIGNAL(desktopAppended(int)),
                rootObject, SIGNAL(qmlDesktopAppended(int)));
        connect(model, SIGNAL(desktopMoved(int, int)),
                rootObject, SIGNAL(qmlDesktopMoved(int, int)));
        connect(model, SIGNAL(currentDesktopChanged(int, int)),
                rootObject, SIGNAL(qmlCurrentDesktopChanged(int, int)));
    }
    else if (isQmlTabletMode)
    {
        connect(model, SIGNAL(rotationModeChanged(QString)),
                rootObject, SIGNAL(qmlOrientationChanged(QString)));
    }

}

QString MultitaskViewManager::getProjectCodeName()
{
    QString projectCodeName = QString(kdk_system_get_projectName());

    return projectCodeName;
}

QString MultitaskViewManager::getProjectSubName()
{
    QString projectSubName = QString(kdk_system_get_projectSubName());

    return projectSubName;
}

QString MultitaskViewManager::getProductFeatures()
{
    QString productFeatures = QString(kdk_system_get_productFeatures());

    return productFeatures;
}

} // namespace MultitaskView
