/*
 * Copyright 2022 KylinSoft Co., Ltd.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

import QtQuick 2.12
import QtQuick.Controls 2.5
import QtQuick.Layouts 1.3
import QtQuick.Window 2.12
import org.kde.kwin 3.0 as KWinComponents
import QtGraphicalEffects 1.0
import "Controller.js" as Controller

FocusScope{
    id : rootDesktopArea
    width : 100
    height : 100
    x : 0
    y : 0

    property ListModel desktopModel : null
    property var rootRect : null
    property var desktopAreaController  : null
    property var keyDesktopController  : null
    property var keyNewDesktopController  : null
    //虚拟桌面的各个参数
    property bool desktopHoverStatus : false
    property double desktopScale : 16/135                                      //虚拟桌面缩放比
    property int desktopThumbnailWidth : rootRect.width * desktopScale        //虚拟桌面缩略图高度
    property int desktopThumbnailHeight : rootRect.height * desktopScale     //虚拟桌面缩略图宽度
    property int desktopCaptionHeight : 24                                  //虚拟桌面标题栏高度
    property int desktopBorderWidth : 4                                    //虚拟桌面边框宽度
    property int desktopRectWidth : desktopThumbnailWidth + 2 * desktopBorderWidth
    property int desktopMoveDistance : (desktopThumbnailWidth + 2 * desktopBorderWidth)/2

    Rectangle {
        id : desktopRect
        color : rootWindow.desktopAreaColor
        width : rootDesktopArea.width
        height : rootDesktopArea.height

        MouseArea {
            id : desktopMouseArea
            anchors.fill : parent
            hoverEnabled : true

            onEntered: {
                desktopHoverStatus =  true
            }

            onExited: {
                desktopHoverStatus = false
            }
        }

        ListView {
            id : desktopListView
            y : desktopRect.height / 2 - desktopThumbnailHeight / 2 - 10
            width : rootDesktopArea.width
            height : desktopRect.height
            property bool initMultitask : true
            orientation : ListView.Horizontal
            model : desktopModel
            spacing : 24
            activeFocusOnTab: false
            Keys.onPressed: {
                if (event.key === Qt.Key_PageUp)
                {
                    if (currentIndex === 0) {
                        return
                    }
                    currentIndex = 0
                    if (desktopListView.contentX > 0){
                        contentMoveAnimation.to = 0
                        contentMoveAnimation.start()
                    }
                }
                else if (event.key === Qt.Key_PageDown)
                {
                    if (currentIndex === desktopModel.count - 2) {
                        return
                    }
                    currentIndex = desktopModel.count - 2
                    if (desktopListView.contentX < desktopListView.contentWidth-desktopListView.width) {
                        contentMoveAnimation.to = desktopListView.contentWidth-desktopListView.width
                        contentMoveAnimation.start()
                    }
                }
                else if (event.key === Qt.Key_Home)
                {
                    if (currentIndex === 0) {
                        return
                    }
                    currentIndex = 0
                    if (desktopListView.contentX > 0){
                        contentMoveAnimation.to = 0
                        contentMoveAnimation.start()
                    }
                }
                else if (event.key === Qt.Key_End)
                {
                    if (currentIndex === desktopModel.count - 2) {
                        return
                    }
                    currentIndex = desktopModel.count - 2
                    if (desktopListView.contentX < desktopListView.contentWidth-desktopListView.width) {
                        contentMoveAnimation.to = desktopListView.contentWidth-desktopListView.width
                        contentMoveAnimation.start()
                    }
                }

            }

            //这里实现了虚拟桌面listview的居中
            NumberAnimation {
                id : desktopListViewLeftMarginAni
                target : desktopListView
                property : "leftMargin"
                duration : 400
                easing.type : Easing.InOutQuad
            }

            add : Transition {
                NumberAnimation {
                    properties : "scale,opacity"
                    from : 0
                    to : 1
                    duration : 400
                    easing.type : Easing.InOutQuad
                }
            }

            remove : Transition{
                ParallelAnimation{
                    NumberAnimation {
                        properties : "scale,opacity"
                        from : 1
                        to : 0
                        duration : 400
                        easing.type : Easing.InOutQuad
                    }
                }
            }

            removeDisplaced : Transition {
                NumberAnimation { properties : "x,y"; duration : 400 }
            }

            move : Transition {
                NumberAnimation {
                    property : "x"
                    duration : 200
                    easing.type : Easing.InOutQuad
                }
            }
            moveDisplaced : Transition {
                NumberAnimation {
                    property : "x"
                    duration : 200
                    easing.type : Easing.InOutQuad
                }
            }

            onCountChanged : {
                if (initMultitask) {
                    if (count === desktopModel.count) {
                        initMultitask = false;
                    }
                    return;
                }
                var newWidth = desktopRectWidth * (count) + desktopListView.spacing * (count)
                desktopListViewLeftMarginAni.to = (parent.width - newWidth) / 2 < 0 ? 0  : (parent.width - newWidth) / 2
                desktopListViewLeftMarginAni.start()
            }

            function bestDelegate(type) {
                if(type === "newDesktop")
                    return "NewDesktopButton.qml"
                return "VirtualDesktopWindow.qml"
            }

            NumberAnimation {
                id : moveToScrollEndAnimation
                target : desktopListView
                property : "contentX"
                from : desktopListView.contentX
                to : desktopListView.contentWidth - desktopListView.width +
                     desktopThumbnailWidth + 2 * desktopBorderWidth + 24
                duration : 300
                easing.type : Easing.InOutQuart
            }

            function moveToScrollEnd() {
                if (desktopListView.contentWidth + desktopThumbnailWidth + 2 * desktopBorderWidth + 24 > desktopListView.width
                     &&  desktopModel.get(desktopListView.count-1).itemType === "newDesktop") {
                    moveToScrollEndAnimation.restart()
                }
            }

            Component.onCompleted: {
                allroot.qmlDesktopAppended.connect(moveToScrollEnd)
            }

            delegate: Loader {
                source : desktopListView.bestDelegate(desktopModel.get(index).itemType)
            }
        } // desktopListView

        NumberAnimation {
            id : contentMoveAnimation
            target : desktopListView
            property : "contentX"
            from : desktopListView.contentX
            to : desktopListView.contentX + desktopMoveDistance
            duration : 300
            easing.type : Easing.InOutQuart
        }

        ArrowButton{
            id : leftButton
            visible : desktopHoverStatus && desktopListView.contentX > 0
            x : 48
            y : (parent.height-leftButton.height)/2
            direction: "left"
        }

        ArrowButton{
            id : rightButton
            visible : desktopHoverStatus && desktopListView.contentX < desktopListView.contentWidth-desktopListView.width
            x : parent.width-48-rightButton.width
            y : (parent.height-rightButton.height)/2
            direction: "right"
        }
    }
}
