/*
 * Copyright 2022 KylinSoft Co., Ltd.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/*
    虚拟桌面的UI和功能的处理，也是加载创建预览窗口
    虚拟桌面区域的窗口和窗口区域的窗口都来自一个"类”
*/
import QtQuick 2.9
import QtQuick.Controls 2.5
import QtQuick.Layouts 1.3
import QtQuick.Window 2.12
import org.kde.kwin 3.0 as KWinComponents
import QtGraphicalEffects 1.0
import "Controller.js" as Controller

FocusScope{
    id : rootDesktopArea
    width : 100
    height : 100
    x : 0
    y : 0
    property ListModel desktopModel : null
    property var rootRect : null
    property var desktopAreaController  : null
    property var keyDesktopController  : null
    property var keyNewDesktopController  : null

    //虚拟桌面的各个参数
    property double desktopScale : 57/480                                    //虚拟桌面缩放比
    property int desktopThumbnailWidth : rootRect.width * desktopScale        //虚拟桌面缩略图高度
    property int desktopThumbnailHeight : rootRect.height * desktopScale     //虚拟桌面缩略图宽度
    property int desktopCaptionHeight : 24                                  //虚拟桌面标题栏高度
    property int desktopBorderWidth : 4                                    //虚拟桌面边框宽度
    property int desktopRectWidth : desktopThumbnailWidth + 2 * desktopBorderWidth
    property bool desktopEnterStatus : false


    Rectangle {
        id : desktopRect
        color : rootWindow.desktopAreaColor
        x : 0
        y : 0
        width : parent.width
        height : parent.height

        MouseArea {
            anchors.fill: parent
            hoverEnabled: true

            onEntered: {
                desktopEnterStatus = true
            }

            onExited: {
                desktopEnterStatus = false
            }

            ListView {
                id : desktopListView
                x : parent.width / 2 - desktopThumbnailWidth / 2 - desktopBorderWidth
                width : parent.width - x
                height : desktopRect.height
                topMargin : 24
                property bool initMultitask : true
                orientation : ListView.Vertical
                model : desktopModel
                spacing : 24
                footerPositioning : ListView.InlineFooter
                activeFocusOnTab: false

                Keys.onPressed: {
                    if (event.key === Qt.Key_PageUp)
                    {
                        if (currentIndex === 0) {
                            return
                        }

                        currentIndex = 0
                        if (desktopListView.contentY > 0){
                            contentMoveAnimation.to = 0
                            contentMoveAnimation.start()
                        }
                    }
                    else if (event.key === Qt.Key_PageDown)
                    {
                        if (currentIndex === desktopModel.count - 2) {
                            return
                        }

                        currentIndex = desktopModel.count - 2
                        if (desktopListView.contentY < desktopListView.contentHeight-desktopListView.height) {
                            contentMoveAnimation.to = desktopListView.contentHeight-desktopListView.height
                            contentMoveAnimation.start()
                        }
                    }
                    else if (event.key === Qt.Key_Home)
                    {
                        if (currentIndex === 0) {
                            return
                        }

                        currentIndex = 0
                        if (desktopListView.contentY > 0){
                            contentMoveAnimation.to = 0
                            contentMoveAnimation.start()
                        }
                    }
                    else if (event.key === Qt.Key_End)
                    {
                        if (currentIndex === desktopModel.count - 2) {
                            return
                        }

                        currentIndex = desktopModel.count - 2
                        if (desktopListView.contentY < desktopListView.contentHeight-desktopListView.height) {
                            contentMoveAnimation.to = desktopListView.contentHeight-desktopListView.height
                            contentMoveAnimation.start()
                        }
                    }

                }

                ScrollBar.vertical: ScrollBar {
                    id : deskVbar
                    hoverEnabled : true
                    active : hovered || pressed

                    //定义样式
                    contentItem : Rectangle {
                        implicitWidth : 8
                        radius : 4
                        visible : desktopEnterStatus && desktopListView.contentHeight > desktopListView.height
                        color : deskVbar.hovered || deskVbar.pressed ? rootWindow.scrollHoverColor : rootWindow.scrollDefaultColor
                    }
                }

                NumberAnimation {
                    id : moveToScrollEndAnimation
                    target : desktopListView
                    property : "contentY"
                    from : desktopListView.contentY
                    to : desktopListView.contentHeight - desktopListView.height +
                         desktopThumbnailHeight + desktopCaptionHeight + 2 * desktopBorderWidth + 2 + 24
                    duration : 300
                    easing.type : Easing.InOutQuart
                }

                function moveToScrollEnd() {
                    if (desktopListView.contentHeight + desktopThumbnailHeight + desktopCaptionHeight + 2 * desktopBorderWidth + 2 + 24 > desktopListView.height
                            && desktopModel.get(desktopListView.count-1).itemType === "newDesktop") {
                        moveToScrollEndAnimation.restart()
                    }
                }

                Component.onCompleted: {
                    allroot.qmlDesktopAppended.connect(moveToScrollEnd)
                }


                add : Transition {
                    NumberAnimation {
                        properties : "scale,opacity"
                        from : 0
                        to : 1
                        duration : 300
                        easing.type : Easing.InOutQuad
                    }
                }

                remove : Transition{
                    ParallelAnimation{
                        NumberAnimation {
                            properties : "scale,opacity"
                            from : 1
                            to : 0
                            duration : 300
                            easing.type : Easing.InOutQuad
                        }
                    }
                }

                removeDisplaced : Transition {
                    NumberAnimation { properties : "x,y"; duration : 300 }
                }

                move : Transition {
                    NumberAnimation {
                        property : "y"
                        duration : 200
                        easing.type : Easing.InOutQuad
                    }
                }
                moveDisplaced : Transition {
                    NumberAnimation {
                        property : "y"
                        duration : 200
                        easing.type : Easing.InOutQuad
                    }
                }

                function bestDelegate(type) {
                    if(type === "newDesktop")
                        return "MasterNewDesktopButton.qml"
                    return "MasterVirtualDesktopWindow.qml"
                }

                delegate: Loader {
                    source : desktopListView.bestDelegate(desktopModel.get(index).itemType)
                }
            } // desktopListView

            NumberAnimation {
                id : contentMoveAnimation
                target : desktopListView
                property : "contentY"
                from : desktopListView.contentY
                to : desktopListView.contentY
                duration : 300
                easing.type : Easing.InOutQuart
            }

        }

    }

}
